/* find_threat.c - find contraband in a pair of CDS images */

/* ---------------------------------------

        _______  _______   _______ 
       /       \/       \ /       \
      /   _____/   __   |/   _____/
     /   /    /   / /   /   /
    /   /    /   / /   / \  \__ 
    |  |___ /   /_/   /__/     \
    |       \        /         /
    \_______/_______/\________/
 
       ----- TRIUMF -----


--------------------------------------- */


/* -------------------------------------------------------------------------------------------------

find_threat:	detect contraband in CDS image data

	Author:
		J.E. Boyd

	Usage:
		find_threat [-o output] [-iso_dens #] [-iso_pdens #] [-bulk on/off] [-sheet on/off]
			[-nrsize #] [-rsize #] [-nt #] [-octa on/off] [-tnrmin #] [-tnrmax #]
			[-trmin #] [-trmax #] [-nr #] [-r #] image-param-file

		-o output specifies output file name base (default "contraband")
		-iso_dens # specifies size of isotropic filter for density data (default 2)
		-iso_pdens # specifies size of isotropic filter for partial density data (default 2)
		-bulk on/off specifies bulk contraband detection algorithm (default on)
		-sheet on/off specifies sheet contraband detection algorithm (default off)
			if both -bulk on and -sheet on, only bulk runs
		-nrsize # specifies size of detection filter for non-resonant density data (default 2)
		-rsize # specifies size of detection filter for resonant partial density data (default 2)
		-nt # specifies number of faces in tessellation of hemishpere to compute sheet directions (default 10)
		-octa on/off specifies octagonal tessellation base instead of standard icosahedron (default off)
		-tnrmin # specifies lower threshold for contraband density data (default 1.3)
		-tnrmax # specifies upper threshold for contraband density data (default 1000)
		-trmin # specifies lower threshold for contraband partial density data (default 0.25)
		-trmax # specifies upper threshold for contraband partial density data (default 1000)
		-nr # specifies non-resonant linear attenuation of matter (cm^2/g) (default 0.025)
		-r # specifies resonant linear attenuation of matter (cm^2/g) (default 0.033)
		image-param-file = image parameter file name

	Inputs:
		stdin:			names of CDS image files
		sinogram files:		CDF image files produced by "recon_image"
					produced from a CDS scan - in files named in stdin
		image parameters:	a set of parameters describing CDS images,
					in the iformat file, image-param-file

	Outputs:
		stdout:			nothing
		image files:		CDF files containing the processed images.
					outname-dens.cdf - isotropically smoothed density
						data (see iso_dens switch)
					outname-pdens.cdf - isotropically smoothed partial
						density data (see iso_pdens switch)

	Description:
		"find_threat" takes a pair of reconstructed CDS images, one non-resonant and
		the other of-resonance, and searches for contraband.  Specifically, it does
		the following:

		1.  it combines and scales the pair of images to give a density image and
		    a partial density image of the CDS FOV

		2.  it isotropically smooths both the density and partial density images and
		    saves them for reference purposes in visualization software

		3.  it applies either a bulk contraband or sheet contraband algorithm to
		    detect the presence of explosives or drugs.

		4.  it analyzes the cluster of contraband voxels produced by either the
		    bulk or sheet algorithm.

		For bulk contraband, "find_threat" isotropically smooths the images and locates
 		contraband voxels in the data using thresholds that define the range of contraband
		densities.  For sheet contraband, "find_threat" applies many smoothing filters to
		isolate sheets in arbitrary orientations.  At each orientation, it locates contraband
		voxels based on the specified thresholds.  It ORs the locations of the contraband
		over the set of directions analzed.  It then anlyzes the cluster produced by the
		ORing over directions.

------------------------------------------------------------------------------------------------- */





#include <stdio.h>
#include <math.h>
#include <cdf.h>
#include "array.h"
#include "getargs.h"
#include "pparse.h"


/* logical constants */
#define		YES		-1
#define		NO		0



/* utility definitions */
#define		MAX( x, y )	( ((x) > (y)) ? (x) : (y) )
#define		ABS(x)		( ((x) >= 0.0) ? (x) : -(x) )
#define		TODEGREES(x)	( 180.0 * (x) / M_PI )


static void CDFstatusHandler( );
static void SaveCDF( );


/* command line switch specifications */

/*command line switch variables */
char *outname = "contraband";	/* base file name for output images */
int iso_dens_size = 2;		/* size of isotropic density filter */
int iso_pdens_size = 2;		/* size of isotropic partial density filter */
int bulk = NO;			/* do bulk material test */
int sheet = YES;		/* do sheet material test */
int nrsize = 2;			/* non-resonant data filter size - half width - filter will be (2 * fltsize + 1)^3 elements */
int rsize = 2;			/* resonant data filter size - half width - filter will be (2 * fltsize + 1)^3 elements */
int ntess = 10;			/* number of tessellation tiles it will take to cover a hemisphere of directions */
int octahedron = NO;		/* use an octahedral basis for tessellation of the unit sphere */
float tnrmin = 1.3;		/* lower nr detection threshold (g/cm^3) */
float tnrmax = 1000.0;		/* upper nr detection threshold (g/cm^3) */
float trmin = 0.25;		/* lower nr detection threshold (g/cm^3) */
float trmax = 1000.0;		/* upper r detection threshold (g/cm^3) */
float nratten = 0.025;		/* non-resonant linear attenuation (cm^2/g) */
float ratten = 0.033;		/* non-resonant linear attenuation (cm^2/g) */


/* command line switch description table */
ARG arglist[] = {
	{	"o",		STRING,		(int *)&outname,	"base name for output images"				},
	{	"iso_dens",	INTEGER,	(int *)&iso_dens_size,	"half size of isotropic density image smoother"		},
	{	"iso_pdens",	INTEGER,	(int *)&iso_pdens_size,	"half size of isotropic partial density image smoother"	},
	{	"bulk",		BOOLEAN,	(int *)&bulk,		"test for bulk matter"					},
	{	"sheet",	BOOLEAN,	(int *)&sheet,		"test for sheet matter"					},
	{	"nrsize",	INTEGER,	(int *)&nrsize,		"non-resonant data filter size (half width)"		},
	{	"rsize",	INTEGER,	(int *)&rsize,		"resonant data filter size (half width)"		},
	{	"nt",		INTEGER,	(int *)&ntess,		"number of tessellations in hemisphere"			},
	{	"octa",		BOOLEAN,	(int *)&octahedron,	"use octahedron for spherical tessellation"		},
	{	"tnrmin",	FLOAT,		(int *)&tnrmin,		"lower non-resonant detection threshold"		},
	{	"tnrmax",	FLOAT,		(int *)&tnrmax,		"upper non-resonant detection threshold"		},
	{	"trmin",	FLOAT,		(int *)&trmin,		"lower resonant detection threshold"			},
	{	"trmax",	FLOAT,		(int *)&trmax,		"upper resonant detection threshold"			},
	{	"nr",		FLOAT,		(int *)&nratten,	"non-resonant linear attenuation (cm^2/g)"		},
	{	"r",		FLOAT,		(int *)&ratten,		"resonant linear attenuation (cm^2/g)"			},
};

#define		NARGS		(sizeof( arglist ) / sizeof( ARG ))



/* image parameter specification */

/* image parameters and default values */
float dfov = 143.0;	/* (cm) diameter of FOV */
float zmotion0 = -50.0;	/* (cm) lower extent of field of view */
float zmotion1 = 50.0;	/* (cm) upper extent of field of view */
int nphi = 180;		/* number of angles to span 180 degrees */
int nl = 287;		/* number of spatial bins across diameter of image */
int nslices = 201;	/* number of slices to span [h0,h1] */
int nenergies = 2;	/* number of gamma energies */
 
/* image parameter description table */
PParseTableEntry ImageParameterParseTable[] = {
	{	"df",		PP_FLOAT,	(int*)&dfov	},
	{	"h0",		PP_FLOAT,	(int*)&zmotion0	},
	{	"h1",		PP_FLOAT,	(int*)&zmotion1	},
	{	"na",		PP_INTEGER,	&nphi		},
	{	"ns",		PP_INTEGER,	&nl		},
	{	"nz",		PP_INTEGER,	&nslices	},
	{	"ne",		PP_INTEGER,	&nenergies	},
};

#define		NImageParameterParseTable	(sizeof(ImageParameterParseTable)/sizeof(PParseTableEntry))



/* maximum length of a file name */
#define			MAXFNAME		200

/* enumeration of bases for tessellation */
#define			OCTAHEDRON		0
#define			ICOSAHEDRON		1



/*  ---------------------------------------  main --------------------------------------- */

main( argc, argv )
int argc;
char *argv[];
{
char fname[MAXFNAME];			/* file names */
float ***nrimage, ***rimage;		/* non-resonant and resonant image data */
float ***nrsmooth, ***rsmooth;		/* smoothed non-resonant and resonant image data */
float ***temp;				/* temporary image data */
float **norm;				/* normals to smoothing planes */
float ***restoration;			/* restored contraband image */
int nnorm;				/* number of smoothing planes */
float TessellateSphere();		/* actual max angle between adjacent planes */
int i;


	/* command line arguments */
	argc = getargs( argc, argv, arglist, NARGS );

	if ( argc != 2 ) {
		fprintf( stderr, "Usage:  %s [switches] image-param-file\n", argv[0] );
		pr_usage( arglist, NARGS );
		exit( NO );
	}


	/* parse image paramter file */
	pparse( argv[1], ImageParameterParseTable, NImageParameterParseTable );

	/* assume that you have images for two energies */
	if ( nenergies != 2 ) {
		fprintf( stderr, "%s:  error in image parameters \"ne\" must equal 2.\n" );
		exit( NO );
	}

	/* allocate image memory */
	nrimage  = Array_3D_Float( 0, nslices-1, 0, nl-1, 0, nl-1 );
	rimage   = Array_3D_Float( 0, nslices-1, 0, nl-1, 0, nl-1 );
	nrsmooth = Array_3D_Float( 0, nslices-1, 0, nl-1, 0, nl-1 );
	rsmooth  = Array_3D_Float( 0, nslices-1, 0, nl-1, 0, nl-1 );
	temp     = Array_3D_Float( 0, nslices-1, 0, nl-1, 0, nl-1 );
	restoration = Array_3D_Float( 0, nslices-1, 0, nl-1, 0, nl-1 );


	/* read image file names and image files */
	ReadFileName( stdin, fname, MAXFNAME );
	fprintf( stderr, "%s: reading <%s>\n", argv[0], fname );
	ReadImageCDF( fname, rimage, nslices, nl, nl );
	ReadFileName( stdin, fname, MAXFNAME );
	fprintf( stderr, "%s: reading <%s>\n", argv[0], fname );
	ReadImageCDF( fname, nrimage, nslices, nl, nl );

	/* subtract nr data from r data and scale to density (g/cm^3) */
	ConditionData( nrimage, rimage, nslices, nl, nl, nratten, ratten );


	/* compute isotropically smoothed density and partial density images and save */
	fprintf( stderr, "%s: isotropic smoothing:\r", argv[0] ); fflush( stderr );
	BulkFilter( nrimage, nrsmooth, temp, nslices, nl, nl, iso_dens_size );
	BulkFilter( rimage, rsmooth, temp, nslices, nl, nl, iso_pdens_size );
	fprintf( stderr, "\n" );
	sprintf( fname, "%s-dens", outname );
	fprintf( stderr, "%s: saving %s\n", argv[0], fname );
	SaveCDF( fname, nrsmooth, nslices, nl, nl );
	sprintf( fname, "%s-pdens", outname );
	fprintf( stderr, "%s: saving %s\n", argv[0], fname );
	SaveCDF( fname,  rsmooth,  nslices, nl, nl );


	/* initialize restoration of contraband to all zeroes */
	InitRestoration( restoration, nslices, nl, nl );


	if ( bulk ) {	/* process for bulk contraband */
		fprintf( stderr, "%s: testing for bulk contraband\r", argv[0] );  fflush( stderr );

		/* isotropic filter of data */
		BulkFilter( nrimage, nrsmooth, temp, nslices, nl, nl, nrsize );
		BulkFilter( rimage, rsmooth, temp, nslices, nl, nl, rsize );

		/* count up the contraband voxels */
		SetContraband( nrsmooth, rsmooth, restoration, nslices, nl, nl, tnrmin, tnrmax, trmin, trmax );
		fprintf( stderr, "\n" );

		/* analyze the cluster */
		AnalyzeCluster( nrimage, rimage, restoration, nslices, nl, nl, tnrmin, tnrmax );
	}
	else
	if ( sheet ) {	/* process for sheet contraband */
		fprintf( stderr, "%s: testing for sheet contraband\n", argv[0] );  fflush( stderr );

		/* compute directions to filter */
		TessellateSphere( ntess * 2, &norm, &nnorm, octahedron ? OCTAHEDRON : ICOSAHEDRON );
		fprintf( stderr, "%s: number of directions: %d\n", argv[0], nnorm );

		/* smooth in each orientation - find number of contraband voxels in each */
		for( i = 0; i < nnorm; i++ ) {
			/* fprintf( stderr, "Direction:  %10.4f %10.4f %10.4f\n", norm[i][1], norm[i][2], norm[i][3] ); */
			fprintf( stderr, "%s: sheet filter: %4da\r", argv[0], i + 1 ); fflush( stderr );
			SheetFilter( nrimage, nrsmooth, temp, nslices, nl, nl, nrsize, norm[i] );
			fprintf( stderr, "%s: sheet filter: %4db\r", argv[0], i + 1 ); fflush( stderr );
			SheetFilter( rimage, rsmooth, temp, nslices, nl, nl, rsize, norm[i] );

			SetContraband( nrsmooth, rsmooth, restoration, nslices, nl, nl, tnrmin, tnrmax, trmin, trmax );
		}
		fprintf( stderr, "\n" );

		/* do cluster analysis on best guess */
		AnalyzeCluster( nrimage, rimage, restoration, nslices, nl, nl, tnrmin, tnrmax );
	}


	/* save restored contraband image */
	sprintf( fname, "%s-ctbnd", outname );
	fprintf( stderr, "%s: saving %s\n", argv[0], fname );
	SaveCDF( fname, restoration, nslices, nl, nl );

	exit( NO );
}

/*  ----------------------------------- end main ---------------------------------------- */




/*  ------------------------------------------------------------------------------------- */
/* InitRestoration - initialize a restoration image to all zeros */

InitRestoration( im, nsl, nr, nc )
float ***im;		/* restoration image */
int nsl, nr, nc;	/* dimension sizes of restoration image */
{
int i, j, k;

	for( i = 0; i < nsl; i++ )
		for( j = 0; j < nr; j++ )
			for( k = 0; k < nc; k++ )
				im[i][j][k] = 0.0;
}




/*  ------------------------------------------------------------------------------------- */
/* AnalyzeCluster - analyze a restoration image - no. of contraband voxels,
             average densisty, partial density, center of gravity, Karhunen-Loeve
             transform of cluster (i.e., enclosing elipsoid) */

/* minimum number of voxels for useful analysis */
#define		MINVOXELS		5

AnalyzeCluster( im1, im2, rst, nsl, nr, nc, tmin1, tmax1 )
float ***im1, ***im2, ***rst;		/* density, partial density, and restored image */
int nsl, nr, nc;			/* image dimension sizes */
float tmin1, tmax1;			/* density threshold */
{
int i, j, k;				/* loop indices */
int n;					/* number of contraband voxels */
float sumx, sumy, sumz;			/* sum of x, y, z postions */
float sumxx, sumyy, sumzz;		/* sum of squares of x, y, z postions */
float sumxy, sumyz, sumxz;		/* sum of x, y, z position products */
float sum1, sum2, sum1sq, sum2sq;	/* sum of density and partial density, sum of squares */
float mean1, mean2, var1, var2;		/* mean density and partial density, variances */
float **a, *d, **v;			/* matrices for Karhunen-Loeve transform */
int nrot;				/* number of rotation required to do Karhunen-Loeve tx */


	/* allocate arrays for Karhunen-Loeve tx */
	a = Array_2D_Float( 1, 3, 1, 3 );
	d = Array_1D_Float( 1, 3 );
	v = Array_2D_Float( 1, 3, 1, 3 );

	/* count up stats on contraband voxels */
	n = 0;
	sumx = sumy = sumz = sumxx = sumyy = sumzz = sumxy = sumyz = sumxz = 0.0;
	sum1 = sum2 = sum1sq = sum2sq = 0.0;
	for( i = 0; i < nsl; i++ )
		for( j = 0; j < nr; j++ )
			for( k = 0; k < nc; k++ )
				if ( tmin1 <= rst[i][j][k] && rst[i][j][k] <= tmax1 ) {
					n++;

					sumx += i;
					sumy += j;
					sumz += k;
					sumxx += i * i;
					sumyy += j * j;
					sumzz += k * k;
					sumxy += i * j;
					sumyz += j * k;
					sumxz += i * k;

					sum1 += im1[i][j][k];
					sum1sq += im1[i][j][k] * im1[i][j][k];
					sum2 += im2[i][j][k];
					sum2sq += im2[i][j][k] * im2[i][j][k];
				}

	/* compute density means and variances */
	mean1 = sum1 / (float)n;
	var1 = sum1sq / (float)n - mean1 * mean1;
	mean2 = sum2 / (float)n;
	var2 = sum2sq / (float)n - mean2 * mean2;

	/* print out analysis results */
	fprintf( stderr, "Analysis:\n" );
	fprintf( stderr, "---------\n" );

	fprintf( stderr, "Number of Voxels: %d\n", n );
	fprintf( stderr, "Center: %10.4f %10.4f %10.4f\n\n", sumx / (float)n, sumy / (float)n, sumz / (float)n );

	fprintf( stderr, "Mean Density:         %10.4f(%g)\n", mean1, sqrt( var1 ) );
	fprintf( stderr, "Mean Partial Density: %10.4f(%g)\n", mean2, sqrt( var2 ) );

	/* if sufficient voxels, do a Karhunen-Loeve transform */
	if ( n > MINVOXELS ) {
		a[1][1] = sumxx / (float)n - sumx * sumx / (float)(n * n);
		a[1][2] = sumxy / (float)n - sumx * sumy / (float)(n * n);
		a[1][3] = sumxz / (float)n - sumx * sumz / (float)(n * n);

		a[2][1] = a[1][2];
		a[2][2] = sumyy / (float)n - sumy * sumy / (float)(n * n);
		a[2][3] = sumyz / (float)n - sumy * sumz / (float)(n * n);

		a[3][1] = a[1][3];
		a[3][2] = a[2][3];
		a[3][3] = sumzz / (float)n - sumz * sumz / (float)(n * n);

		jacobi( a, 3, d, v, &nrot );

		/* fprintf( stderr, "\n = %d\nNumber of Rotations: %d\n", n, nrot ); */

		fprintf( stderr, "Karhunen-Loeve Transform:\n" );
		fprintf( stderr, "  Eigenvalues:   %10.4f %10.4f %10.4f\n", d[1], d[2], d[3] );
		fprintf( stderr, "  Eigenvectors:  %10.4f %10.4f %10.4f\n", v[1][1], v[1][2], v[1][3] );
		fprintf( stderr, "                 %10.4f %10.4f %10.4f\n", v[2][1], v[2][2], v[2][3] );
		fprintf( stderr, "                 %10.4f %10.4f %10.4f\n\n", v[3][1], v[3][2], v[3][3] );

	} else {
		fprintf( stderr, "Analyze: too few voxels for Karhunen-Loeve analysis\n" );
	}


	Free_2D_Float( a, 1, 3, 1, 3 );
	Free_1D_Float( d, 1, 3 );
	Free_2D_Float( v, 1, 3, 1, 3 );

}






/*  ------------------------------------------------------------------------------------- */
/* SetContraband - set contraband voxels in restoration to max of previous restoration
                   and density indicated in density image */

SetContraband( im1, im2, rest, nsl, nr, nc, tmin1, tmax1, tmin2, tmax2 )
float ***im1, ***im2, ***rest;	/* density, partial density, and restoration images */
int nsl, nr, nc;		/* image dimension sizes */
float tmin1, tmax1;		/* density thresholds */
float tmin2, tmax2;		/* partial density thresholds */
{
int i, j, k;	/* loop indices */

	for( i = 0; i < nsl; i++ )
		for( j = 0; j < nr; j++ )
			for( k = 0; k < nc; k++ )
				/* test thresholds to find contraband */
				if ( tmin1 <= im1[i][j][k] && im1[i][j][k] <= tmax1 &&
						tmin2 <= im2[i][j][k] && im2[i][j][k] <= tmax2 ) {

					/* if contraband set restoration to max or restoration and density */
					rest[i][j][k] = MAX( rest[i][j][k], im1[i][j][k] );
				}
}



/*  ------------------------------------------------------------------------------------- */
/* ConditionData - convert a non-resonant and on-resonance image to density
                   and partial density images */

ConditionData( nrim, rim, nsl, nr, nc, nratt, ratt )
float ***nrim, ***rim;		/* non-resonant and on-resonance images */
int nsl, nr, nc;		/* image dimension sizes */
float nratt, ratt;		/* linear attenuations for density and partial density */
{
int i, j, k;	/* loop indices */

	for( i = 0; i < nsl; i++ )
		for( j = 0; j < nr; j++ )
			for( k = 0; k < nc; k++ ) {
				/* subtract nr component from r image */
				rim[i][j][k] -= nrim[i][j][k];

				/* scale images to density */
				nrim[i][j][k] /= nratt;
				rim[i][j][k] /= ratt;
			}
}




/*  ------------------------------------------------------------------------------------- */
/* ReadFileName - read the name of a file from input stream */

ReadFileName( fp, name, max )
FILE *fp;		/* input stream to read from */
char *name;		/* character string for file name */
int max;		/* maximum length of string */
{
int len;

	fgets( name, max, fp );		/* read file name (including \n if enough room in "name") */
	if ( feof( fp ) ) {		/* test to see if eof occurred - exit if so */
		fprintf( stderr,
			"find_threat:  unexpected eof while reading sinogram file name\n" );
		exit( YES );
	}

	/* test to see if \n is in name - if so, replace it with \0 */
	len = strlen( name );
	if ( name[len-1] == '\n' ) name[len-1] = '\000';
}



/*  ------------------------------------------------------------------------------------- */
/* ReadImageCDF - read a CDF image */

ReadImageCDF( name, image, nsl, nl )
char *name;		/* CDF file name */
float ***image;		/* three dimensional image */
int nsl;		/* number of slices in image */
int nl;			/* number of spatial steps in each dimension of a slice */
{
CDFid id;			/* id of CDF when open */
CDFstatus status;		/* return status variable for CDF calls */
void CDFstatusHandler();	/* CDF status handler */
int slicevar;			/* variable number for slices in CDF */
int slice;			/* loop index for reading slices */
int index[2], count[2], interval[2];


	/* open the CDF */
	status = CDFopen( name, &id );
	if ( status != CDF_OK ) CDFstatusHandler( status );
	slicevar = CDFvarNum( id, "slice" );

	/* read the CDF */
	index[1] = 0;
	count[0] = 1;
	count[1] = nl;
	interval[0] = interval[1] = 1;
	for( slice = 0; slice < nsl; slice++ ) {
		for( index[0] = 0; index[0] < nl; index[0]++ ) {
			status = CDFvarHyperGet( id, slicevar, slice, 1, 1, index, count, interval, image[slice][index[0]] );
			if ( status != CDF_OK ) CDFstatusHandler( status );
		}
	}

	CDFclose( id );
}


/*  ------------------------------------------------------------------------------------- */
/* SaveCDF - create a CDF image */

static void SaveCDF( fname, im, nsl, nr, nc )
char *fname;		/* name of CDF to create */
float ***im;		/* 3D  image */
int nsl;		/* number of slices */
int nr, nc;		/* number of rows and columns */
{
CDFstatus status;			/* return status of CDF calls */
CDFid id;				/* id of CDF */
long var;				/* variable in CDF */
long dimsizes[2], dimvarys[2];		/* dimension sizes and variabilities */
long index[2], count[2], interval[2];	/* dimension indices, counts and intervals */
long slice, l, phi;			/* slice number, l index and phi index */
static void CDFstatusHandler();		/* CDF status error handler */

	/* test to see if CDF already exists, if so delete it */
	status = CDFopen( fname, &id );
	if ( status >= CDF_OK ) {
		status = CDFdelete( id );
		if ( status != CDF_OK ) CDFstatusHandler( status );
	}

	/* create a new CDF with variable the size of a sinogram slice */
	dimsizes[0] = nr;
	dimsizes[1] = nc;
	status = CDFcreate( fname, 2L, dimsizes, NETWORK_ENCODING, ROW_MAJOR, &id );
	if ( status != CDF_OK ) CDFstatusHandler( status );

	/* create variables for data */
	dimvarys[0] = dimvarys[1] = VARY;
	status = CDFvarCreate( id, "slice", CDF_FLOAT, 1L, VARY, dimvarys, &var );
	if ( status != CDF_OK ) CDFstatusHandler( status );

	/* write the data to the CDF */
	index[0] = index[1] = 0;
	count[0] = nr; count[1] = nc;
	interval[0] = interval[1] = 1;
	status = CDFvarHyperPut( id, var,   0, nsl, 1L, index, count, interval, &(im[0][0][0]) );
	if ( status != CDF_OK ) CDFstatusHandler( status );

	/* close CDF to make sure it is properly written */
	status = CDFclose( id );
	if ( status != CDF_OK ) CDFstatusHandler( status );
}


/*  ------------------------------------------------------------------------------------- */
/* CDFstatusHandler - handle CDF error status codes */

static void CDFstatusHandler( status )
CDFstatus status;
{
char message[CDF_ERRTEXT_LEN+1];

	if ( status < CDF_WARN ) {
		fprintf( stderr, "An error has occurred, halting...\n" );
		CDFerror( status, message );
		fprintf( stderr, "%s\n", message );
		exit( status );
	} else if ( status < CDF_OK ) {
		fprintf( stderr, "Warning:  function may not have completed as expected...\n" );
		CDFerror( status, message );
		fprintf( stderr, "%s\n", message );
	} else if ( status > CDF_OK ) {
		fprintf( stderr, "Function completed successfully, but advised that...\n" );
		CDFerror( status, message );
		fprintf( stderr, "%s\n", message );
	}
	return;
}


